import json
from abc import ABC, abstractmethod
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.model_selection import train_test_split
from sklearn.naive_bayes import MultinomialNB
from sklearn import metrics
import copy
from simbots.utils.exceptions import SchemaException

class Intent(ABC):
    """
    The abstract method for any intent.
    """

    @abstractmethod
    def __init__(self, intentName="Irrelevant",allIntentExamples=None, entityHandler=None,testSize=0.25,vocab=None):

        self.intentName = intentName
        self.testSize = testSize

        if not allIntentExamples:
            allIntentExamples = {}

        self.entityHandler = entityHandler
        self.allIntentExamples = allIntentExamples
        if self.entityHandler:
            self.substituteAllEntitiesInAllTrainingExamples()

        self.classifier = None
        self.vocab=vocab

        self.trainedIntentArray = []
        self.trained = False
        self.createIntent()

    def substituteAllEntitiesInAllTrainingExamples(self):

        """

        For certain types of classifiers it is helpful to have entities substituted
        Whenever a new intent is trained , all the entitiy values in the message texts get replaced by the entity kind values.
        This is to help training the intents . Note that you can define certain entities to help increase the accuracy of intent


        """

        allIntentExamplesSubstituted = {}
        allIntentExamplesCopy = copy.deepcopy(self.allIntentExamples)

        for key in allIntentExamplesCopy:
            dum = []
            for mes in allIntentExamplesCopy[key]:
                dum.append(self.entityHandler.substituteAllEntities(" {0} ".format(mes)))

            allIntentExamplesSubstituted[key] = dum

        self.allIntentExamples = allIntentExamplesSubstituted

    @abstractmethod
    def createIntent(self,clfName=None):
        """

        Method to create intent will be overridden in the derived class

        :param clfName: Classifier Name
        :return: the inherited class should return a classifier

        """
        pass

    @staticmethod
    @abstractmethod
    def train(X_train, y_train, params = None):
        """

        :param X_train: Training examples
        :param y_train: Training labels
        :param params: Any additional parameters that may be needed can be defined here
        :return:

        """
        pass


class MultinomialNBIntent(Intent):
    """
    Uses Multinomial Naive bayes classifier for intent training.
    """

    def __init__(self,intentName="Irrelevant",allIntentExamples=None,entityHandler = None,vocab=None):
        """

        :param allIntentExamples:
        :param entityHandler:
        """

        super().__init__(intentName=intentName,allIntentExamples=allIntentExamples,
                         entityHandler = entityHandler,testSize=0.25,vocab=vocab)


    def substituteAllEntitiesInAllTrainingExamples(self):
        """
        No changes made to

        :return:
        """
        super().substituteAllEntitiesInAllTrainingExamples()


    def createVocab(self):
        """
        This function creates the vocab for intent and updates the intent vocabulary
        :return:
        count_vect : This returns a fitted Countvectorizer object
        """

        ##
        ## Initialise the Count Vectoriser
        ##
        count_vect = CountVectorizer() # stop_words="english"
        ##
        ## Create the Count Vectoriser Object (This gives a bag of words representation)
        ##
        allIntentExamplesCopy = copy.deepcopy(self.allIntentExamples)

        allData=[intentExample for intentExamples in allIntentExamplesCopy.values() for intentExample in intentExamples]
        count_vect.fit_transform(allData)
        vocab=count_vect.vocabulary_
        count_vect.revVocabulary_ =  dict(zip(vocab.values(),vocab.keys()))
        self.vocab = count_vect

        return count_vect

    def createDataSetForIntent(self):
        """
        Creates the data set for the intent model
        :return:
        X_train: training sentences asn list
        X_test: testing sentences as list
        X : all sentences as list
        y_train: train set labels
        y_test: testset labels
        y : the complete  sample labels (including the training and the testing sets)

        """
        ##
        ## if output is 1 means intent is present else not
        ##
        allIntentExamplesCopy = copy.deepcopy(self.allIntentExamples)

        intentSamples = allIntentExamplesCopy[self.intentName]

        allIntentExamplesCopy.pop(self.intentName, None)
        nonIntentSamples = [intentExample for intentExamples in allIntentExamplesCopy.values() for intentExample in
                            intentExamples]

        Y = [1 for i in range(len(intentSamples))] + [0 for i in range(len(nonIntentSamples))]
        intentSamples.extend(nonIntentSamples)
        X = copy.deepcopy(intentSamples)

        ##
        ## stratify split the dataset to maintain class balance
        ##
        X_train, X_test, y_train, y_test = train_test_split(X, Y, test_size=self.testSize, stratify=Y, random_state=2020)

        return X_train, X_test, X, y_train, y_test, Y

    @staticmethod
    def train(X_train, y_train, params = None):
        """
        Train the intent model

        :param X_train: Training sentences
        :param y_train: Training labels
        :param params: None
        :return: trained classifier
        """
        vectorizer = params["vectorizer"]
        X_trainCount = vectorizer.transform(X_train)
        clf = MultinomialNB()
        clf.fit(X_trainCount, y_train)
        return clf

    @staticmethod
    def predictClassifierOutput(clf, X, count_vect):
        """
        Make predictions on the trained intent
        :param clf: trained classifier
        :param X: The sample for prediction
        :param count_vect: count vectoriser for the classifier
        :return:

        predictions: predicted class
        probabilities: probablities for the predicted class
        """
        X_count = count_vect.transform(X)
        predictions = clf.predict(X_count)
        probabilities = clf.predict_proba(X_count)
        return predictions, probabilities


    def getIntentProbability(self, userMessage):
        """
        Get the intent probability
        :param userMessage: Text message from the user
        :return: {
        "name": classifier name
        "confidence": classifier confidence
        }
        """
        userMessage = [userMessage]
        predictions, probabilities = self.predictClassifierOutput(self.classifier["classifier"], userMessage,
                                                             self.classifier["data"]["countVect"])


        return {"name": self.classifier["name"], "confidence": probabilities[0][1]}

    def createIntent(self, clfName=None):
        """

        :param clfName: Classifier Name , if not provided , will default to intentName+'_multinomialNbClassifier'
        :return: classifier

        classifier = {
            "name": Classifier name,
            "classifier": Classifier Object,
            "expectedAccuracy": Training accuracy achieved by the intent,
            "data": {
                "countVect": count vectoriser ,
                "trainedOnX": Training samples,
                "trainedOnY": Training labels
            }



        """
        if not clfName:
            clfName = self.intentName + "_multinomialNbClassifier"

        ##
        ## Create vocabulary
        ##

        if not self.vocab:
            cv = self.createVocab()
        else:
            cv= self.vocab

        X_train, X_test, X_all, y_train, y_test, y_all = self.createDataSetForIntent()

        clf = self.train(X_train, y_train, {"vectorizer" :cv})

        predictions, probabilities = self.predictClassifierOutput(clf, X_test, cv)

        expectedAccuracy = metrics.accuracy_score(y_test, predictions)

        ##
        ## Now train on the complete dataset
        ##
        clf = self.train(X_all, y_all, {"vectorizer" :cv})

        classifier = {
            "name": clfName,
            "classifier": clf,
            "expectedAccuracy": expectedAccuracy,
            "data": {
                "countVect": cv,
                "trainedOnX": X_train,
                "trainedOnY": y_train
            }
        }

        self.classifier = classifier
        return classifier



class IntentsHandler():
    """
    Class for creating and handling multiple intents
    """

    def __init__(self,allIntentExamples=None, entityHandler=None):

        """

        Intent Schema validation

        """

        if type(allIntentExamples)  != dict:
            raise SchemaException("Intent Samples","",
                                  "Intents should be a dict of the schema : \n "+
                                  "{\n 'IntentName1' : list of str  ,\n 'IntentName2' : list of str \n "+"}"

                                  )

        for intentName in allIntentExamples.keys():
            intentExamples = allIntentExamples[intentName]

            if type(intentExamples) != list:
                raise SchemaException("Intent",intentName ,"Intent should be a list of type str , instead intent is {0} .".format(type(intentExamples)))

            for i,intentSample in enumerate(intentExamples):

                if type(intentSample) != str:
                    raise SchemaException("Intent",intentName,"Intent should be a list of type str ,instead intent sample no {0} ({1}) is {2} .".format(i,intentSample,type(intentSample)))

        intentNames = allIntentExamples.keys()
        if "Irrelevant" not in intentNames:
            raise SchemaException("Intent Samples","No intent with the name Irrelevant .","Every bot should have an intent with the name 'Irrelevant' , not found in detected intent names .")
        ##
        ## Initialise intent
        ##
        self.allIntentExamples = allIntentExamples
        self.intentSamplesAugment()
        self.entityHandler     = entityHandler
        self.trained    = False
        self.createAllTrainedIntents()


    def intentSamplesAugment(self):
        """

        Function to augment intent samples ,making sure each intent is atleast 100 samples long


        """

        for key in self.allIntentExamples:
            while len(self.allIntentExamples[key]) < 100:
                self.allIntentExamples[key].extend(self.allIntentExamples[key])




    def createAllTrainedIntents(self):
        """
        Creates trained intents and sets self.trained

        """

        allIntentExamplesCopy = copy.deepcopy(self.allIntentExamples)

        trainedIntentArray = []
        intentNames  = list(allIntentExamplesCopy.keys())
        intentNames.sort()
        #trainedIntentArray.append(MultinomialNBIntent(intentName=intentNames[0],allIntentExamples=self.allIntentExamples,entityHandler=self.entityHandler))
        #vocab = trainedIntentArray[0].vocab

        for intentName in intentNames:
            #
            # allIntentExamplesCopy = copy.deepcopy(self.allIntentExamples)
            #
            trainedIntentArray.append(
                MultinomialNBIntent(intentName=intentName,allIntentExamples=self.allIntentExamples,entityHandler=self.entityHandler,vocab=None)

            )
        self.trainedIntentArray =trainedIntentArray
        self.trained=True



    def getMessageIntents(self, message,dialogNumber=0):
        """

        :param message: Message Text
        :param dialogNumber:  The current dialog number
        :return:
        list of [{
        "name": classifier name
        "confidence": Classifier confidence
        "dialogNumber": Current dialog number
        "rank": Classifier rank
        }]
        """

        if self.entityHandler:
            # substitute message entities if they are available
            message = self.entityHandler.substituteAllEntities(message)

        if self.trained:
            allIntents = sorted([trainedIntent.getIntentProbability(message) for trainedIntent in self.trainedIntentArray],
                      key=lambda x: x["confidence"], reverse=True)

            allIntentsDialogCounterUpdated=[]
            rank=1
            for elem in allIntents:
                elem["dialogNumber"]=dialogNumber
                elem["rank"]=rank
                allIntentsDialogCounterUpdated.append(elem)
                rank+=1

            return allIntentsDialogCounterUpdated
        else:
            return "model not trained"






from .Intent import IntentsHandler
from .Entity import EntitiesHandler
from objectpath import Tree
import datetime

class ContextManager():
    """
    Context handler class, This maintains context as a dict and also as a Object Path tree, to facilitate searching within the context
    """

    def __init__(self,entitiesExtractorJson,allIntentExamples):

        entitiesHandler =  EntitiesHandler(entitiesExtractorJson)
        intentHandlers  = [IntentsHandler(allIntentExamples,entityHandler=entitiesHandler)]

        self.entitiesHandler = entitiesHandler
        self.intentHandlers  = intentHandlers
        self.context={
            "entities":[],
            "intents":[],
            "dialogs":[],
            "messages":[]
        }
        self.contextTree = Tree(self.context)

    def setNewContext(self,newContext):
        """
        Deletes the old context and sets a new one .
        :param newContext: the new context object as json

        """
        self.context = newContext
        self.contextTree =Tree(self.context)

    def clearContext(self):
        """
        Clears context
        """
        context = {
            "entities": [],
            "intents": [],
            "dialogs": [],
            "messages": []
        }
        self.setNewContext(context)



    def updateDialog(self,message,type="input"):
        """
        Updates the context by
        1) Computing all entities present in the message
        2) Extracting all intents in the message
        3) Updating the current dialog number
        4) Updating messages in the context


        :param message: Message text
        :param type: message type : "input" or "output"

        """

        if type =="input":
            # for input
            dialogNumber = len(self.context["dialogs"]) + 1
            extractedEntities = self.entitiesHandler.extractAllEntities(message,dialogNumber=dialogNumber)
            extractedIntents  = [ intent for ihandler in self.intentHandlers for intent in ihandler.getMessageIntents(message=message,dialogNumber=dialogNumber) ]

            self.context["entities"].extend(extractedEntities)
            self.context["intents"].extend(extractedIntents)
            self.context["dialogs"].append(dialogNumber)
            messageJson={
                "text":message,
                "entitiesSub":self.entitiesHandler.substituteAllEntities(message),
                "messageType":type,
                "dialogNumber":dialogNumber,
                "sendTime":  datetime.datetime.now().strftime("%d/%m/%Y, %H:%M:%S")

            }
        else:
            # for output message
            dialogNumber = len(self.context["dialogs"])
            messageJson ={
                "text":message,
                "entitiesSub":message,
                "messageType":type,
                "dialogNumber":dialogNumber,
                "sendTime": datetime.datetime.now().strftime("%d/%m/%Y, %H:%M:%S")

            }

        self.context["messages"].append(messageJson)
        # update the tree
        self.contextTree =Tree(self.context)

    def findStuff(self,filter=None,stuff="intents"):
        """
        Basic method to search a particular part of the context ie Intent, Entities, Message
        :param filter: dict of terms to search
        :param stuff: part of context that needs to be searched ie Intent, Entities, Message
        :return: matched values as list
        """
        matchedStuff=[]
        if filter:
            # $.intss[@.foo is 1][@.bar is b]
            query=["$.{0}".format(stuff)]
            for key in filter:
                if key != "confidence":
                    query.append("[@.{0} is {1}]".format(key,filter[key]))
                else:
                    query.append("[@.{0} > {1}]".format(key,filter[key]))

            query="".join(query)

            matchedStuff =list(self.contextTree.execute(query))
        return matchedStuff

    def findCurrentTopIntent(self):
        """
        Find the the top intent in the current message
        :return: topIntent: the Top intent ie intent with rank 1
        """
        # Find current dialog number
        currentDialogNumber = self.context["dialogs"][-1]

        # Find
        #if confidenceLimit:
        topIntent= list(self.findStuff({"dialogNumber": currentDialogNumber, "rank": 1}, stuff="intents"))

        #print "topfound",topIntent
        if len(topIntent)>0:
            topIntent=topIntent[-1]
        else:
            topIntent={}

        return topIntent

    def findCurrentEntities(self):
        """
        Find the entities detected in the current message
        :return: currentEntities : the entities detected in the current message as list
        """
        # Find current dialog number
        currentDialogNumber = self.context["dialogs"][-1]

        # Find current Entities
        currentEntities = list(self.findStuff({"dialogNumber": currentDialogNumber}, stuff="entities"))


        return currentEntities

    @staticmethod
    def findObjectInStuff(stuff,filter):
        """
        Generalised method to search within any given dictionary using objectPath library
        :param stuff: the dictionary where the object is to be searched
        :param filter: the filter
        :return: list of matched objects
        """

        allStuff={
            "allStuff":stuff,

        }
        stuffTree=Tree(allStuff)
        query=["$.allStuff"]
        for key in filter:
            query.append("[@.{0} is {1}]".format(key, filter[key]))

        query = "".join(query)
        return list(stuffTree.execute(query))






from objectpath import Tree
import random
from .Context import ContextManager

class Bot():
    """
    Whenever a new bot is created it will inherit from the Bot class.

    """

    def __init__(self, intentExamples, entityExamples, templates, confidenceLimit=0):
        """

        :param intentExamples: The Intents need to be in a json format as accepted by IntentsHandler
        :param entityExamples: The Entities need to be in a json format as accepted by entities handler
        :param templates:  Bot output message templates
        :param confidenceLimit: Confidence limit for any intent to be accepted to generate a response
        """

        contextManager = ContextManager(allIntentExamples=intentExamples, entitiesExtractorJson=entityExamples)
        self.contextManager = contextManager
        self.templatesTree = Tree(templates)
        self.templates = templates
        self.confidenceLimit = confidenceLimit

    def updateConversationInputs(self, message):
        """
        Update the conversation by updating the context with the input message
        :param message:
        :return:
        """
        self.contextManager.updateDialog(message)

    def updateConversationOutputs(self,message):
        """
        Update the conversation by updating the context with the output message
        :param message:
        :return:
        """
        self.contextManager.updateDialog(message,"output")


    def reason(self):
        """
        Bot reasoning is done here , use bot intents and entities to create message logic here

        :return:
        Should return a list of
        {
        "tag": Used to identify output message from the template
        "data": To be used to fill the missing attributes in message to generate the complete message
        }

        """

        ## find current dialogNumber

        currentDialogNumber = self.contextManager.context["dialogs"][-1]

        currentTopIntent = self.contextManager.findCurrentTopIntent()

        currentEntities = self.contextManager.findCurrentEntities()

        output = []

        if (currentTopIntent["confidence"] < self.confidenceLimit) or (currentTopIntent["name"] == "Irrelevant"):
            currentTopIntent = {}

        if currentTopIntent:
            ##
            ## Rule For All
            ##
            # if currentTopIntent["name"]=="Greetings":
            ##
            ## Person is greeting
            ##
            name = currentTopIntent["name"].split("_")[0]
            reply = {
                "tag": "{0}.basic".format(name),
                "data": None,
                # "message":""

            }
            output.append(reply)

        else:
            ##
            ##
            ## Rule for irrelevant
            ##
            ##
            irrelevant = {
                "tag": "Irrelevant.basic",
                "data": None

            }
            output.append(irrelevant)

        return output

    def say(self, output, outputTheme):
        """
        Generate bot output message by adding message to the output components
        :param output:
        :param outputTheme:
        :return:
        """

        botSpeech = []

        for component in output:
            componentTag = component["tag"]
            query = "$.{0}.{1}".format(outputTheme, componentTag)
            queryResults = self.templatesTree.execute(query)
            message = random.choice(queryResults)
            data = component["data"]
            if data:
                message = message.format(data)


            botSpeech.append(message)



        return botSpeech

    def getBotContext(self):
        """
        Returns bot context
        :return: context as json
        """
        return self.contextManager.context

    def getBotConfidence(self):
        """
        Returns the bot confidence as list of tuples
        :return: [(intentName,intentConfidence) ..]
        """
        currentDialogNumber = self.contextManager.context["dialogs"][-1]
        return [(elem["name"], elem["confidence"]) for elem in
                self.contextManager.findStuff(filter={"dialogNumber": currentDialogNumber}, stuff="intents")]

    def getBotOutput(self, message, outputTheme):
        # update the conversation with Inputs
        self.updateConversationInputs(message)

        # reason
        output = self.reason()

        # say
        messages = self.say(output, outputTheme)

        # update the conversation with Outputs
        for message in messages:
            self.updateConversationOutputs(message)



        return "\n".join(messages)





class EntitySamples():
    """
    Contains samples for frequently used entities

    """

    @staticmethod
    def greetingsHelper():
        """
        :return:

        {'wsup': [{'tag': 'case-insensitive',
                   'pattern': "\s[w]*[a]*[s]+[u]+[p]+\s",
                   'type': 'regex'}],
         'hi': [{'tag': 'case-insensitive',
                 'pattern': "\s[h]+[i]+\s", 'type': 'regex'}],
         'hello': [{'tag': 'case-insensitive',
                    'pattern': "\s[h]+[e]+[l]+[o]+\s",
                    'type': 'regex'}]}



        """
        return {'wsup': [{'tag': 'case-insensitive',
                   'pattern': "\s[w]*[a]*[s]+[u]+[p]+\s",
                   'type': 'regex'}],
         'hi': [{'tag': 'case-insensitive',
                 'pattern': "\s[h]+[i]+\s", 'type': 'regex'}],
         'hello': [{'tag': 'case-insensitive',
                    'pattern': "\s[h]+[e]+[l]+[o]+\s",
                    'type': 'regex'}]}

    @staticmethod
    def laughterHelper():
        """
        :return:

        {'haha': [{'tag': 'case-insensitive','pattern': "\s(h+(a|e)+)+(h+)?\s",'type': 'regex'}],'happysmily': [{'tag': 'case-insensitive','pattern': "\s\:\)\s", 'type': 'regex'}]}
        """

        return {'haha': [{'tag': 'case-insensitive',
                                 'pattern': "\s(h+(a|e)+)+(h+)?\s",
                                 'type': 'regex'}],
                       'happysmily': [{'tag': 'case-insensitive',
                                       'pattern': "\s\:\)\s", 'type': 'regex'}]}

    @staticmethod
    def coolHelper():
        """
        :return:

        {'cool': [{'tag': 'case-insensitive','pattern': "\sc+oo+l+\s", 'type': 'regex'}]}

        """
        return {'cool': [{'tag': 'case-insensitive','pattern': "\sc+oo+l+\s", 'type': 'regex'}]}

    @staticmethod
    def byeHelper():
        """
        :return:
        {'bye': [{'tag': 'case-insensitive','pattern': "\s(goo+d)?b+y+e+\s", 'type': 'regex'}]}

        """
        return {'bye': [{'tag': 'case-insensitive','pattern': "\s(goo+d)?(b+u+)?b+y+e+\s", 'type': 'regex'}]}


#from examples import entities as entitiesExtractorJson
import re
import json
from simbots.utils.exceptions import SchemaException

class EntitiesHandler():
    """
    This class is responsible for handling entities within the bot
    The init method for the EntitiesHandler Class
        :param entitiesExtractorJson: this is a json which specifies the type of entities and their definitions . Note currently no schema checks have been implemented .

        **Sample Json accepted**

        .. code-block:: json

            {
                    "GreetingsHelper":{
                        "wsup" : [{
                                    "tag":"case-insensitive",
                                    "pattern":"\s[w]*[a]*[s]+[u]+[p]+\s",
                                    "type":"regex"
                        }
                                ],
                        "hi":[
                            {
                                "tag":"case-insensitive",
                                "pattern":"\s[h]+[i]+\s",
                                "type":"regex"


                            }

                        ],
                        "hello":[
                            {
                                "tag":"case-insensitive",
                                "pattern":"\s[h]+[e]+[l]+[o]+\s",
                                "type":"regex"


                            },
                            {
                               "tag":"case-insensitive",
                               "extractor":helloExtractorFunc ->
                               "type":"function"

                            }

                        ]


                    },

                    "LaughterHelper":{

                        "haha" : [{
                                    "tag":"case-insensitive",
                                    "pattern":"\s(h+(a|e)+)+(h+)?\s",
                                    "type":"regex"


                        }
                                ],
                        "happysmily":[
                            {
                                "tag":"case-insensitive",
                                "pattern":"\s\:\)\s",
                                "type":"regex"


                            }

                        ],

                    },
                    "CoolHelper":{
                        "cool":[
                            {
                                "tag":"case-insensitive",
                                "pattern":"\sc+oo+l+\s",
                                "type":"regex"

                            }


                        ]


                    },
                    "ByeHelper":{
                        "bye":[
                            {
                                "tag":"case-insensitive",
                                "pattern":"\s(goo+d)?b+y+e+\s",
                                "type":"regex"

                            }
                        ]

                    }

                }

        **kind** : Here GreetingsHelper,LaughterHelper,CoolHelper,ByeHelper are entity kinds.
        also "wsup","hi","hello" are entity values assosciated with the entity kind GreetingsHelper

        **type**: Only two types are supported -> regex and function.

        **pattern** : the pattern of the regex, only used if type is regex,

        **extractor**: A function that extracts the given entity from the message text should be present if type is function,

        Note that any function used should take in two arguments messageText and dialogNumber and should
        return A list of dicts of the format

        .. code-block:: json

            [{
                        "value":the value of the entity , for more details look at the sample json.,
                        "exactValue":the exact value that occurs within the message text.,
                        "kind":the kind of the entity eg  food in case of pizza.,
                        "dialogNumber":the dialog number of the current message -> same as the input param.,
                        "foundAt":[start location , end location],

            }]


        **substituter**: A custom substitution function to substitute entity by another value, This function takes in the message text and should
        return the substituted message text this should always be present if type is "function". should return substituted message text.


    """

    def __init__(self,entitiesExtractorJson=None):
        if not entitiesExtractorJson:
            entitiesExtractorJson={}
        else:
            # validate entity schema here
            if type(entitiesExtractorJson) != dict:
                # check if entity is dict
                raise SchemaException("Entities", "Entity samples",
                                      "Entities should be specified as a dict instead got  {0} .".format(type(entitiesExtractorJson)))

            for entityName in entitiesExtractorJson.keys():
                entityContent = entitiesExtractorJson[entityName]
                if type(entityContent) != dict:
                    raise SchemaException("Entity", entityName,
                                          "Entities should be specified as a dict instead got  {0} .".format(
                                              type(entityContent)))
                for entitySynonym in entityContent.keys():
                    synonymContent = entityContent[entitySynonym]
                    if type(synonymContent) != list:
                       raise SchemaException("Entity","{0} - {1}".format(entityName,entitySynonym),"Should be a list of dicts instead got {0} of {1}".format(synonymContent,type(synonymContent))

                                             )
                    for i,synonymDefinition in enumerate(synonymContent):
                        if type(synonymDefinition) != dict:
                            raise SchemaException("Entity","{0} - {1}- index {2}".format(entityName,entitySynonym,i),"Should be a dict got {0} of {1}".format(synonymDefinition,type(synonymDefinition))

                                             )
                        for key in synonymDefinition.keys():
                            if key == 'tag':
                                value = synonymDefinition[key]
                                if value not in ["case-insensitive","case-sensitive"]:
                                    raise SchemaException("Entity","{0} - {1}- index {2}".format(entityName,entitySynonym,i),"allowed 'tag' values in ['case-insensitive','case-sensitive'] got {0}".format(value))
                            if key == "type":
                                value = synonymDefinition[key]
                                if value not in ["regex", "function"]:
                                    raise SchemaException("Entity",
                                                          "{0} - {1}- index {2}".format(entityName, entitySynonym,
                                                                                         i),
                                                          "allowed 'type' values in ['regex', 'function'] got {0}".format(
                                                              value))
                            if key == "pattern":
                                value = synonymDefinition[key]
                                if type(value) != str:
                                    raise SchemaException("Entity",
                                                          "{0} - {1}- index {2}".format(entityName, entitySynonym,
                                                                                         i),
                                                          "pattern should be a regex got {0}".format(
                                                              type(value)))
                            if key =="extractor":
                                # work more on this
                                value = synonymDefinition[key]
                                if not callable(value):
                                    raise SchemaException("Entity",
                                                          "{0} - {1}- index {2}".format(entityName, entitySynonym,
                                                                                         i),
                                                          "extractor should be a function got {0}".format(
                                                              type(value)))
                                # write extractor text here

                            if "tag" not in synonymDefinition.keys():
                                raise SchemaException("Entity",
                                                      "{0} - {1}- index {2}".format(entityName, entitySynonym, i),
                                                      "should have  a key called 'tag' ")
                            if "type" not in synonymDefinition.keys():
                                raise SchemaException("Entity",
                                                      "{0} - {1}- index {2}".format(entityName, entitySynonym, i),
                                                      "should have  a key called 'type' ")

                            if ("tag" =='regex') and "pattern" not in synonymDefinition.keys():
                                raise SchemaException("Entity",
                                                      "{0} - {1}- index {2}".format(entityName, entitySynonym, i),
                                                      "if  'tag' =='regex' you must add a key called 'pattern' with a regex value  ")

                            if ("tag" =='function') and "extractor" not in synonymDefinition.keys():
                                raise SchemaException("Entity",
                                                      "{0} - {1}- index {2}".format(entityName, entitySynonym, i),
                                                      "if  'tag' =='function' you must add a key called 'extractor' with a function that extracts the entities and a key called 'substituter' which is a function that substitues the message entities for more details refer documentation -> EntitiesHandler")

                            if ("tag" =='function') and "substituter" not in synonymDefinition.keys():
                                raise SchemaException("Entity",
                                                      "{0} - {1}- index {2}".format(entityName, entitySynonym, i),
                                                      "if  'tag' =='function' you must add a key called 'extractor' with a function that extracts the entities and a key called 'substituter' which is a function that substitues the message entities for more details refer documentation -> EntitiesHandler")





        self.entitiesExtractorJson=entitiesExtractorJson


    def extractEntitySynonym(self,message,pattern,value,kind,tag,dialogNumber):
        """
            This function extracts the entities and synonyms from the text message and returns a json. Currently it does not extract any synonyms
                :param message: The text message str eg "I want to buy a pizza"
                :param pattern:
                :param value:
                :param kind:
                :param tag: can be "case-insensitive" if we want the extractor to ignore the casing (uppercase or lowercase) of the entity text
                :param dialogNumber: the dialog number of the message string.
                :return: A list of dicts of the format

                .. code-block:: json

                    [{
                                "value":the value of the entity , for more details look at the sample json.,
                                "exactValue":the exact value that occurs within the message text.,
                                "kind":the kind of the entity eg : food in case of pizza.,
                                "dialogNumber":the dialog number of the current message -> same as the input param.,
                                "foundAt":[start location , end location],

                    }]
        """


        entitiesFound=[]
        if "case-insensitive" in tag:
            iter=re.finditer(pattern=pattern,string= message,flags=re.I)
        else:
            iter = re.finditer(pattern=pattern, string=message)

        for entityDetected in iter:

            exactValue=message[entityDetected.start() : entityDetected.end()]
            entitiesFound.append(
                {
                    "value":value,
                    "exactValue":exactValue.strip(),
                    "kind":kind,
                    "dialogNumber":dialogNumber,
                    "foundAt":[entityDetected.start() , entityDetected.end()],
                    #"logic":synonymSpecs

                }
            )
        return entitiesFound


    def extractAllEntities(self,message,dialogNumber=0):
        """
        This function extracts all entities from the message .
            :param message: The message text
            :param dialogNumber:  the current dialog number
            :return: A list of dicts of the format

            .. code-block:: json

                [{
                            "value":the value of the entity , for more details look at the sample json.,
                            "exactValue":the exact value that occurs within the message text.,
                            "kind":the kind of the entity eg : food in case of pizza.,
                            "dialogNumber":the dialog number of the current message -> same as the input param.,
                            "foundAt":[start location , end location],

                }]
        """
        #message = " {0} ".format(message)
        allEntities=[]
        for entityKind in  self.entitiesExtractorJson.keys():
            for entityValue in self.entitiesExtractorJson[entityKind].keys():
                for synonymSpecs in self.entitiesExtractorJson[entityKind][entityValue]:
                    if synonymSpecs["type"] == "regex":
                        tag=synonymSpecs["tag"]
                        synonymPattern=synonymSpecs["pattern"]
                        entFound =  self.extractEntitySynonym(message,synonymPattern,entityValue,entityKind,tag,dialogNumber)
                    elif synonymSpecs["type"] == "function":
                        entFound = synonymSpecs["extractor"](message,dialogNumber)
                    else:
                        entFound =[]

                    allEntities.extend(entFound)


        return allEntities



    def substituteEntityKind(self,message,pattern,by,tag):
        """
        This function substitutes an entity value for its kind in the message text and returns the substituted message text .
        Note this function is currently unstable and a better version needs to be written to substitute entities effectively .
            :param message: The message text
            :param pattern: the entity pattern
            :param by:  the value to be substituted by eg if we need to substitute the word "equivalent" by "alike" then by ="alike"
            or if pizza then it will be kind
            :param tag:"case-insensitive" or not
            :return: the substituted message text
        """

        #message = " {0} ".format(message)
        if "case-insensitive" in tag:
            flags=re.I
        else:
            flags=None

        messageVar = re.sub( pattern , by, message,flags=flags)

        return messageVar

    def substituteAllEntities(self,message):
        """
        This function substitutes all entity values with entity kinds
            :param message:The message text
            :return: substituted message text
        """
        #message =" {0} ".format(message)
        for entityKind in  self.entitiesExtractorJson.keys():
            for entityValue in self.entitiesExtractorJson[entityKind].keys():
                for synonymSpecs in self.entitiesExtractorJson[entityKind][entityValue]:
                    if synonymSpecs["type"] == "regex":
                        tag=synonymSpecs["tag"]
                        synonymPattern=synonymSpecs["pattern"]
                        message = self.substituteEntityKind(message, synonymPattern, " --!{0} ".format(entityKind), tag)
                    elif synonymSpecs["type"] == "function":
                        message = synonymSpecs["substituter"](message)


        return message





















from setuptools import setup


from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name = "simbots",
    version = "0.0.17",
    description = "Simple bots or Simbots is a library designed to create simple bots using the power of python.",
    packages =["simbots","simbots\\utils"],
    package_dir = {'':'simbots'},
    classifiers = ["Development Status :: 1 - Planning","Operating System :: OS Independent","License :: OSI Approved :: MIT License","Programming Language :: Python :: 3"],
    install_requires =['sklearn','objectpath'],
    license_files = ('LICENSE.txt'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    author = "Vaibhav Arora",
    author_email ="vaibhav.rr1@gmail.com"


)



class IntentSamples():
    """
    Contains samples for frequently used intents, which can be used to train intents .Each method returns a list of samples

    """
    def __init__(self):
        pass

    @staticmethod
    def greetingSamples():
        """
        :return: list of samples like  'Hello !', 'Hi , How are you ?',
        """
        return [
        'Hello !',
        'Hi , How are you ?',
        'Hey !! ',
        'Namaste ',
        'Good day',
        'Good evening',
        'Good morning',
        'Good to see you',
        'Greetings',
        'Have you been well?',
        'Hello Agent',
        'Hello',
        'Hey how are you doing',
        'Hey there all',
        'Hey there',
        'Hey twin',
        'Hey you',
        'Hi advisor',
        'Hi there',
        'How are things going?',
        'How are you today?',
        'How have you been?',
        'How is it going?',
        'How r u?',
        'Looking good eve',
        "What's new?",
        "What's up?",
        'You there',
        'Namaste',
        'satsriyakaal',
        'helo',
        'hiiiiiii',
        ' sup ',
        ' wassup bruh ',
        ' sup bro ',
        ' ssssup mate whats up',
            "hi",
            "hey"
    ]
    @staticmethod
    def botNameSamples():
        """
        :return:  list of samples like  'Who is this?', 'Hi' ,'Who are you',
        """
        return [
        'Who is this?',
        'Who are you',
        'Who are you ?',
        'What is your Name ?',
        'What is your name ',
        'what should i call you ?',
        'What should i call you',
        'How to address you ?',
        'how  to address you by name',
        'what is your name ',
        'call you',
        'your name',
        'call you ?',
        'your name ?',
        'whom are you',
        'whom are you referred as',
    ]

    @staticmethod
    def relativeSamples():
        """

        :return:  list of samples like  'Do you have any relatives ' , "do you have any siblings ?"
        """
        return [
        'Do you have any relatives ',
        'Do you have a sibling ',
        'Any brother or sister',
        'Do you have a father',
        'Do you happen to have a grandparent ?',
        'Any grandparents or parents ',
        'do you have Any wife ?',
        'Are you married ?',
    ]

    @staticmethod
    def ageSamples():
        """
        :return:  list of samples like  'what is your age' , "how old are you ?"

        """
        return [
            'What is your age',
            'your age ?',
            'how old are you ?'
            'when were you born',
            'your year of birth',
            'year of birth',
            'when were you born again',
            'born ?',
            'born again ?',
            "how long have you lived ?"

        ]

    @staticmethod
    def birthPlaceSamples():
        """

        :return:  list of samples like  'where were you born ' , "where are you from ?"
        """
        return [
        'where were you born ?',
        'where are you from',
        'where in the world are you from ?',
        'where do you belong to',
        'you are from ',
        'are you from ',
        'you belong to ',
        'place of birth',
        'and you are from ?',
        'and your birthplace is from ?and you are from ?',
        'birthplace',
        'place of birth',
        'location of birth',
        'place of residence',
        'location',
        'your address',
        'where do you reside ?',
        'place of residence ?',
    ]

    @staticmethod
    def abilitiesSamples():
        """

        :return:  list of samples like  'what can you do ? ' , "what are your capabilities ?"
        """
        return [
        'What can you do ?',
        'What are your abilities ',
        'your abilities',
        'your capabilities',
        'what are your capabilities',
        'ability',
        'your ability',
        'capability',
        'your capability',
        'your abilities',
        'your powers',
        'what are your superpowers',
        'superpowers',
        'you are capable of ',
        'what are you capable of',
        'what can you do ?',
        'what can you do ?',
        'Can you jump ?',
        'can you shit ?',
        'could you do this ?',
        'could you please do this ?',
    ]
    @staticmethod
    def reallySamples():
        """

        :return:  list of samples like  'is that so ? ' , "really ?"
        """
        return [
            'really',
            'is that so ?',
            'is that what you believe',
            'are you sure',
            'is that what you belief',
            'you positive about that',
            'are you positively sure',
            'is that what you think',
            'are you positive about that ?',
            'you positive',
            'are you really sure',
            'really bro ?',
            "ya sure ?"
        ]

    @staticmethod
    def laughterSamples():
        """

        :return:  list of samples like  'hahaha' , "thats very funny"
        """
        return [
        'i think thats funny ',
        'thats very funny ,hahaha',
        'i think that hahahaha',
        'thats very funny',
        'laughter',
        ':)',
        'thats really funny ',
        'I think thats hilarious',
        'thats so funny !',
        'that really made me laugh !',
        "you are so funny :)"
    ]


    @staticmethod
    def coolSamples():
        """

        :return:  list of samples like  "thats cool !","cool"
        """
        return [
        'thats cool',
        'cool',
        'that is really cool',
        'very cool',
        'totally cool !',
        'Awesome !',
        'Very cool',
        'Pretty cool',
        'surely very cool',
        'nice',
        'neat',
        'thats neat',
        'thats pretty neat',
        'thats nice',
    ]

    @staticmethod
    def praiseSamples():
        """

        :return:  list of samples like  'you are awesome ' , "thats incredible !"
        """
        return [
        'I think that you are absolutely amazing  !',
        'I really think you are amazing',
        'You are absolutely fantastic',
        'Thats Incredible',
        "you are awesome buddy !",
        "this is awesome ",
        "talking to you is awesome",
        " i believe you are awesome !!",
        "pretty awesome my friend !",
        "thanks for being so awesome !!"
        'How Extraordinary !',
        'Far Out ! ',
        'Great ! ',
        'Outstanding',
        'Performance',
        'Marvelous',
        'I Cant Get Over It !',
        'Wonderful ! ',
        'Amazing Effort !',
        'Unbelievable Work',
        'You Should Be Proud',
        'Phenomenal ! ',
        'Youve Got It',
        'Superb ! ',
        'Youre Special',
        'Excellent ! ',
        'Cool ! ',
        'Your Project Is First Rate !',
        'Way to Go ! ',
        'Youve Outdone',
        'Yourself',
        'Thumbs Up',
        'What A Great',
        'Listener',
        'Your Help Counts ! ',
        'You Came Through ! ',
        'Terrific',
        'You Tried Hard',
        'Youre OK',
        'Fabulous',
        'You Made It',
        'Happen',
        'Youre a Real',
        'Trooper',
        'It Couldnt Be',
        'Better',
        'The Time You Put in Shows',
        'Bravo ! ',
        'Youre Unique',
        'Exceptional',
        'Fantastic Work',
        'Breathtaking ! ',
        'Youre a Great',
        'Example For Others',
        'Keep Up the Good',
        'Work',
        'Awesome  !',
        'I Knew You Had It In You',
        'Youve Made',
        'Progress',
        'Your Work Is Out of Sight',
        'What an Imagination  ! ',
        'Its Everything I Hoped For',
        'Stupendous',
        'Youre Sensational',
        'Very Good !',
    ]

    @staticmethod
    def trueSamples():
        """

        :return:  list of samples like  'thats true' , "thats evident"
        """
        return [
        'Thats True',
        'thats self evident',
        'thats evident',
        'thats truly evident',
        'true that',
        'true as it comes',
        'thats completely true',
        'true as it comes',
        'thats correct',
        'absolutely correct',
    ]

    @staticmethod
    def falseSamples():
        """

        :return:  list of samples like  'thats false ' , "you are wrong "
        """
        return [
        'thats false',
        'you are wrong',
        'completely wrong',
        'absolutely and utterly wrong',
        'I dont believe that',
        'i think thats not correct ',
        'thats absolutely not correct',
    ]

    @staticmethod
    def byeSamples():
        """

        :return:  list of samples like  'bye ' , "goodbye"
        """
        return [
        'good bye',
        'bye bye',
        'buh bye',
        'see you later',
        'byee',
        'se ya later aligator',
        'I gotta go',
        'I have someplace else i need to be',
        'I Need to be someplace else',
        'We can talk later',
        'Will chat later',
        'We can chat later',
        'Will be chatting later',
        'goodbye',
    ]

    @staticmethod
    def confirmSamples():
        """

        :return:  list of samples like  'confirm that' , "yes","do that please"
        """
        return [
            'Yes ',
            'Okay',
            'Please do that',
            'Okay go ahead',
            'You do that',
            'Wrap it up',
            'I confirm',
            'Please do',
            'Sure go ahead',
            'Sure',
            'ok',
            'ok',
            'yep',
            'yep yep do that ',
        ]

    @staticmethod
    def thanksSamples():
        """
        :return:  list of samples like  "Thanks"," Thank you"
        """
        return ['Thank you', 'Thanks', 'Thank you so very much !'," my thanks !"]
    @staticmethod
    def discardSamples():
        """

        :return:  list of samples like  'No","cancel that "
        """

        return [
            'No',
            'No , i dont wanna do that',
            'Nope change that',
            ' negative ',
            'No no',
            'I said no',
            'Never',
            'Cancel that',
            'Build a new one',
        ]

    @staticmethod
    def jokeSamples():
        """

        :return:  list of samples like  "can you tell me a joke","tell a joke"
        """
        return [
        'Can you tell a joke ?',
        'Tell a joke !',
        'Tell me another joke !',
        'I want to smile ',
        'can you make me smile',
        'could you tell me a joke ?',
        'make me laugh',
        'I want some laughter',
        'make some laughter',
        'give me some jokes ',
    ]





class SchemaException(Exception):
    """
    Exception raised for errors in schema.

    """

    def __init__(self, type=None,name=None, message=None):

        self.message = "\n\n Error encountered in {0} : {1} \n Detail : {2}".format(type,name,message)
        super().__init__(self.message)




